@@ -1,5 +1,6 @@
 use std::char;
 use std::str;
+use std::cmp;
 
 use {
     ErrorPos,
@@ -9,7 +10,7 @@ use {
     XmlCharExt,
 };
 
-pub type Result<T> = ::std::result::Result<T, StreamError>;
+type Result<T> = ::std::result::Result<T, StreamError>;
 
 
 /// Representation of the [Reference](https://www.w3.org/TR/xml/#NT-Reference) value.
@@ -26,7 +27,7 @@ pub enum Reference<'a> {
 }
 
 
-/// Streaming text parsing interface.
+/// A streaming text parsing interface.
 #[derive(PartialEq, Clone, Copy, Debug)]
 pub struct Stream<'a> {
     bytes: &'a [u8],
@@ -35,9 +36,19 @@ pub struct Stream<'a> {
     span: StrSpan<'a>,
 }
 
-impl<'a> Stream<'a> {
-    /// Constructs a new `Stream` from a string span.
-    pub fn from_span(span: StrSpan<'a>) -> Stream {
+impl<'a> From<&'a str> for Stream<'a> {
+    fn from(text: &'a str) -> Self {
+        Stream {
+            bytes: text.as_bytes(),
+            pos: 0,
+            end: text.len(),
+            span: text.into(),
+        }
+    }
+}
+
+impl<'a> From<StrSpan<'a>> for Stream<'a> {
+    fn from(span: StrSpan<'a>) -> Self {
         Stream {
             bytes: span.to_str().as_bytes(),
             pos: 0,
@@ -45,17 +56,9 @@ impl<'a> Stream<'a> {
             span,
         }
     }
+}
 
-    /// Constructs a new `Stream` from a string.
-    pub fn from_str(text: &str) -> Stream {
-        Stream {
-            bytes: text.as_bytes(),
-            pos: 0,
-            end: text.len(),
-            span: StrSpan::from_str(text),
-        }
-    }
-
+impl<'a> Stream<'a> {
     /// Returns an underling string span.
     pub fn span(&self) -> StrSpan<'a> {
         self.span
@@ -93,7 +96,7 @@ impl<'a> Stream<'a> {
     ///
     /// # Errors
     ///
-    /// Returns `UnexpectedEndOfStream` if we are at the end of the stream.
+    /// - `UnexpectedEndOfStream`
     pub fn curr_byte(&self) -> Result<u8> {
         if self.at_end() {
             return Err(StreamError::UnexpectedEndOfStream);
@@ -133,10 +136,10 @@ impl<'a> Stream<'a> {
     ///
     /// # Errors
     ///
-    /// Returns `UnexpectedEndOfStream` if we are at the end of the stream.
+    /// - `UnexpectedEndOfStream`
     pub fn next_byte(&self) -> Result<u8> {
         if self.pos + 1 >= self.end {
-            return Err(StreamError::UnexpectedEndOfStream.into());
+            return Err(StreamError::UnexpectedEndOfStream);
         }
 
         Ok(self.bytes[self.pos + 1])
@@ -146,10 +149,10 @@ impl<'a> Stream<'a> {
     ///
     /// # Errors
     ///
-    /// Returns `UnexpectedEndOfStream` if we are at the end of the stream.
+    /// - `UnexpectedEndOfStream`
     pub fn curr_char(&self) -> Result<char> {
         if self.at_end() {
-            return Err(StreamError::UnexpectedEndOfStream.into());
+            return Err(StreamError::UnexpectedEndOfStream);
         }
 
         Ok(self.curr_char_unchecked())
@@ -167,7 +170,7 @@ impl<'a> Stream<'a> {
     /// ```rust,should_panic
     /// use xmlparser::Stream;
     ///
-    /// let mut s = Stream::from_str("text");
+    /// let mut s = Stream::from("text");
     /// s.advance(2); // ok
     /// s.advance(20); // will cause a panic via debug_assert!().
     /// ```
@@ -186,7 +189,7 @@ impl<'a> Stream<'a> {
     /// ```
     /// use xmlparser::Stream;
     ///
-    /// let mut s = Stream::from_str(" \t\n\r &#x20; ");
+    /// let mut s = Stream::from(" \t\n\r &#x20; ");
     /// s.skip_spaces();
     /// assert_eq!(s.at_end(), true);
     /// ```
@@ -238,7 +241,7 @@ impl<'a> Stream<'a> {
     /// ```
     /// use xmlparser::Stream;
     ///
-    /// let mut s = Stream::from_str("Some text.");
+    /// let mut s = Stream::from("Some text.");
     /// s.advance(5);
     /// assert_eq!(s.starts_with(b"text"), true);
     /// assert_eq!(s.starts_with(b"long"), false);
@@ -264,7 +267,7 @@ impl<'a> Stream<'a> {
             is_space = true;
         } else if c == b'&' {
             // Check for (#x20 | #x9 | #xD | #xA).
-            let mut s = self.clone();
+            let mut s = *self;
             if let Some(v) = s.try_consume_char_reference() {
                 if (v as u32) < 255 && (v as u8).is_xml_space() {
                     is_space = true;
@@ -280,11 +283,15 @@ impl<'a> Stream<'a> {
     /// Like [`skip_spaces()`], but checks that first char is actually a space.
     ///
     /// [`skip_spaces()`]: #method.skip_spaces
+    ///
+    /// # Errors
+    ///
+    /// - `InvalidChar`
     pub fn consume_spaces(&mut self) -> Result<()> {
         if !self.at_end() && !self.starts_with_space() {
             let c = self.curr_byte_unchecked() as char;
             let pos = self.gen_error_pos();
-            return Err(StreamError::InvalidChar(c, "Space".into(), pos).into());
+            return Err(StreamError::InvalidSpace(c, pos));
         }
 
         self.skip_spaces();
@@ -296,13 +303,14 @@ impl<'a> Stream<'a> {
     /// # Errors
     ///
     /// - `InvalidChar`
+    /// - `UnexpectedEndOfStream`
     ///
     /// # Examples
     ///
     /// ```
     /// use xmlparser::Stream;
     ///
-    /// let mut s = Stream::from_str("Some text.");
+    /// let mut s = Stream::from("Some text.");
     /// s.consume_byte(b'S').unwrap();
     /// s.consume_byte(b'o').unwrap();
     /// s.consume_byte(b'm').unwrap();
@@ -312,10 +320,9 @@ impl<'a> Stream<'a> {
         if self.curr_byte()? != c {
             return Err(
                 StreamError::InvalidChar(
-                    self.curr_byte_unchecked() as char,
-                    String::from_utf8(vec![c]).unwrap(),
+                    vec![self.curr_byte_unchecked(), c],
                     self.gen_error_pos(),
-                ).into()
+                )
             );
         }
 
@@ -330,13 +337,15 @@ impl<'a> Stream<'a> {
     /// # Errors
     ///
     /// - `InvalidChar`
+    /// - `UnexpectedEndOfStream`
     pub fn consume_either(&mut self, list: &[u8]) -> Result<u8> {
         assert!(!list.is_empty());
 
         let c = self.curr_byte()?;
         if !list.contains(&c) {
-            let s = String::from_utf8(list.to_vec()).unwrap();
-            return Err(StreamError::InvalidChar(c as char, s, self.gen_error_pos()).into());
+            let mut v = list.to_vec();
+            v.insert(0, c);
+            return Err(StreamError::InvalidChar(v, self.gen_error_pos()));
         }
 
         self.advance(1);
@@ -349,13 +358,19 @@ impl<'a> Stream<'a> {
     ///
     /// - `InvalidChar`
     pub fn skip_string(&mut self, text: &[u8]) -> Result<()> {
-        // TODO: use custom error
         if !self.starts_with(text) {
-            for c in text {
-                self.consume_byte(*c)?;
-            }
+            let len = cmp::min(text.len(), self.end - self.pos);
+            // Collect chars and do not slice a string,
+            // because the `len` can be on the char boundary.
+            // Which lead to a panic.
+            let actual = self.span.to_str()[self.pos..].chars().take(len).collect();
 
-            // unreachable
+            // Assume that all input `text` are valid UTF-8 strings, so unwrap is safe.
+            let expected = str::from_utf8(text).unwrap().to_owned();
+
+            let pos = self.gen_error_pos();
+
+            return Err(StreamError::InvalidString(vec![actual, expected], pos));
         }
 
         self.advance(text.len());
@@ -376,7 +391,7 @@ impl<'a> Stream<'a> {
 
         let name = self.slice_back(start);
         if name.is_empty() {
-            return Err(StreamError::InvalidName.into());
+            return Err(StreamError::InvalidName);
         }
 
         Ok(name)
@@ -385,13 +400,18 @@ impl<'a> Stream<'a> {
     /// Skips an XML name.
     ///
     /// The same as `consume_name()`, but does not return a consumed name.
+    ///
+    /// # Errors
+    ///
+    /// - `InvalidNameToken` - if name is empty or starts with an invalid char
+    /// - `UnexpectedEndOfStream`
     pub fn skip_name(&mut self) -> Result<()> {
         let mut iter = self.span.to_str()[self.pos..self.end].chars();
         if let Some(c) = iter.next() {
             if c.is_xml_name_start() {
                 self.advance(c.len_utf8());
             } else {
-                return Err(StreamError::InvalidName.into());
+                return Err(StreamError::InvalidName);
             }
         }
 
@@ -441,11 +461,11 @@ impl<'a> Stream<'a> {
             (prefix, local)
         } else {
             let local = self.slice_back(start);
-            (StrSpan::from(""), local)
+            ("".into(), local)
         };
 
         if local.is_empty() {
-            return Err(StreamError::InvalidName.into());
+            return Err(StreamError::InvalidName);
         }
 
         Ok((prefix, local))
@@ -468,24 +488,25 @@ impl<'a> Stream<'a> {
 
     /// Consumes quote.
     ///
-    /// Consumes ''' or '"' and returns it.
+    /// Consumes `'` or `"` and returns it.
     ///
     /// # Errors
     ///
-    /// - `InvalidChar`
+    /// - `InvalidQuote`
+    /// - `UnexpectedEndOfStream`
     pub fn consume_quote(&mut self) -> Result<u8> {
         let c = self.curr_byte()?;
         if c == b'\'' || c == b'"' {
             self.advance(1);
             Ok(c)
         } else {
-            Err(StreamError::InvalidChar(c as char, "'\"".into(), self.gen_error_pos()).into())
+            Err(StreamError::InvalidQuote(c as char, self.gen_error_pos()))
         }
     }
 
-    /// Consumes bytes by predicate and returns them.
+    /// Consumes bytes by the predicate and returns them.
     ///
-    /// Result can be empty.
+    /// The result can be empty.
     pub fn consume_bytes<F>(&mut self, f: F) -> StrSpan<'a>
         where F: Fn(&Stream, u8) -> bool
     {
@@ -494,7 +515,7 @@ impl<'a> Stream<'a> {
         self.slice_back(start)
     }
 
-    /// Consumes bytes by predicate.
+    /// Consumes bytes by the predicate.
     pub fn skip_bytes<F>(&mut self, f: F)
         where F: Fn(&Stream, u8) -> bool
     {
@@ -508,9 +529,9 @@ impl<'a> Stream<'a> {
         }
     }
 
-    /// Consumes chars by predicate and returns them.
+    /// Consumes chars by the predicate and returns them.
     ///
-    /// Result can be empty.
+    /// The result can be empty.
     pub fn consume_chars<F>(&mut self, f: F) -> StrSpan<'a>
         where F: Fn(&Stream, char) -> bool
     {
@@ -519,7 +540,7 @@ impl<'a> Stream<'a> {
         self.slice_back(start)
     }
 
-    /// Consumes chars by predicate.
+    /// Consumes chars by the predicate.
     pub fn skip_chars<F>(&mut self, f: F)
         where F: Fn(&Stream, char) -> bool
     {
@@ -539,21 +560,29 @@ impl<'a> Stream<'a> {
     pub fn try_consume_char_reference(&mut self) -> Option<char> {
         let start = self.pos();
 
-        match self.consume_reference() {
-            Ok(Reference::CharRef(ch)) => Some(ch),
-            _ => {
-                self.pos = start;
-                None
-            }
+        if let Ok(Reference::CharRef(ch)) = self.consume_reference() {
+            Some(ch)
+        } else {
+            self.pos = start;
+            None
         }
     }
 
     /// Consumes an XML reference.
     ///
     /// Consumes according to: <https://www.w3.org/TR/xml/#NT-Reference>
+    ///
+    /// # Errors
+    ///
+    /// - `InvalidReference`
+    /// - `UnexpectedEndOfStream`
     pub fn consume_reference(&mut self) -> Result<Reference<'a>> {
+        self._consume_reference().map_err(|_| StreamError::InvalidReference)
+    }
+
+    fn _consume_reference(&mut self) -> Result<Reference<'a>> {
         if self.curr_byte()? != b'&' {
-            return Err(StreamError::InvalidReference.into());
+            return Err(StreamError::InvalidReference);
         }
 
         self.advance(1);
@@ -561,25 +590,19 @@ impl<'a> Stream<'a> {
             self.advance(1);
             let n = if self.curr_byte()? == b'x' {
                 self.advance(1);
-                let value = self.consume_bytes(|_, c| c.is_xml_hex_digit());
-                match u32::from_str_radix(value.to_str(), 16) {
-                    Ok(v) => v,
-                    Err(_) => return Err(StreamError::InvalidReference.into()),
-                }
+                let value = self.consume_bytes(|_, c| c.is_xml_hex_digit()).to_str();
+                u32::from_str_radix(value, 16).map_err(|_| StreamError::InvalidReference)
             } else {
-                let value = self.consume_bytes(|_, c| c.is_xml_digit());
-                match u32::from_str_radix(value.to_str(), 10) {
-                    Ok(v) => v,
-                    Err(_) => return Err(StreamError::InvalidReference.into()),
-                }
-            };
+                let value = self.consume_bytes(|_, c| c.is_xml_digit()).to_str();
+                u32::from_str_radix(value, 10).map_err(|_| StreamError::InvalidReference)
+            }?;
 
             let c = char::from_u32(n).unwrap_or('\u{FFFD}');
-            if c.is_xml_char() {
-                Reference::CharRef(c)
-            } else {
-                return Err(StreamError::InvalidReference.into());
+            if !c.is_xml_char() {
+                return Err(StreamError::InvalidReference);
             }
+
+            Reference::CharRef(c)
         } else {
             let name = self.consume_name()?;
             match name.to_str() {
@@ -598,48 +621,50 @@ impl<'a> Stream<'a> {
     }
 
     /// Slices data from `pos` to the current position.
-    pub fn slice_back(&mut self, pos: usize) -> StrSpan<'a> {
+    pub fn slice_back(&self, pos: usize) -> StrSpan<'a> {
         self.span.slice_region(pos, self.pos())
     }
 
     /// Slices data from the current position to the end.
-    pub fn slice_tail(&mut self) -> StrSpan<'a> {
+    pub fn slice_tail(&self) -> StrSpan<'a> {
         self.span.slice_region(self.pos(), self.end)
     }
 
     /// Calculates a current absolute position.
     ///
     /// This operation is very expensive. Use only for errors.
+    #[inline(never)]
     pub fn gen_error_pos(&self) -> ErrorPos {
         let row = self.calc_current_row();
         let col = self.calc_current_col();
         ErrorPos::new(row, col)
     }
 
-    /// Calculates a current absolute position.
+    /// Calculates an absolute position at `pos`.
     ///
     /// This operation is very expensive. Use only for errors.
-    pub fn gen_error_pos_from(&mut self, pos: usize) -> ErrorPos {
-        let old_pos = self.pos;
-        self.pos = pos;
-        let e = self.gen_error_pos();
-        self.pos = old_pos;
+    #[inline(never)]
+    pub fn gen_error_pos_from(&self, pos: usize) -> ErrorPos {
+        let mut s = *self;
+        let old_pos = s.pos;
+        s.pos = pos;
+        let e = s.gen_error_pos();
+        s.pos = old_pos;
         e
     }
 
-    fn calc_current_row(&self) -> usize {
+    fn calc_current_row(&self) -> u32 {
         let text = self.span.full_str();
         let mut row = 1;
         let end = self.pos + self.span.start();
-        row += text.as_bytes()
-            .iter()
-            .take(end)
-            .filter(|c| **c == b'\n')
-            .count();
-        row
+        row += text.bytes()
+                   .take(end)
+                   .filter(|c| *c == b'\n')
+                   .count();
+        row as u32
     }
 
-    fn calc_current_col(&self) -> usize {
+    fn calc_current_col(&self) -> u32 {
         let text = self.span.full_str();
         let bytes = text.as_bytes();
         let end = self.pos + self.span.start();
