@@ -1,22 +1,11 @@
 use std::fmt;
 
-use error::Result;
-use stream::Result as StreamResult;
-use {
-    ElementEnd,
-    EntityDefinition,
-    Error,
-    ExternalId,
-    FromSpan,
-    Stream,
-    StreamError,
-    StrSpan,
-    Token,
-    XmlByteExt,
-    XmlCharExt,
-};
+use super::*;
 
 
+type Result<T> = ::std::result::Result<T, Error>;
+type StreamResult<T> = ::std::result::Result<T, StreamError>;
+
 #[derive(Clone, Copy, PartialEq)]
 enum State {
     Document,
@@ -78,77 +67,63 @@ impl fmt::Display for TokenType {
 }
 
 
-/// Tokenizer of the XML structure.
+/// Tokenizer for the XML structure.
 pub struct Tokenizer<'a> {
     stream: Stream<'a>,
     state: State,
     depth: usize,
+    is_fragment_parsing: bool,
 }
 
-impl<'a> FromSpan<'a> for Tokenizer<'a> {
-    fn from_span(span: StrSpan<'a>) -> Self {
+impl<'a> From<&'a str> for Tokenizer<'a> {
+    fn from(text: &'a str) -> Self {
+        Self::from(StrSpan::from(text))
+    }
+}
+
+impl<'a> From<StrSpan<'a>> for Tokenizer<'a> {
+    fn from(span: StrSpan<'a>) -> Self {
         Tokenizer {
-            stream: Stream::from_span(span),
+            stream: Stream::from(span),
             state: State::Document,
             depth: 0,
+            is_fragment_parsing: false,
         }
     }
 }
 
-impl<'a> Iterator for Tokenizer<'a> {
-    type Item = Result<Token<'a>>;
 
-    fn next(&mut self) -> Option<Self::Item> {
-        if self.stream.at_end() || self.state == State::Finished {
-            self.state = State::Finished;
-            return None;
-        }
-
-        let t = Self::parse_next_impl(&mut self.stream, self.state);
-
-        if let Some(ref t) = t {
-            match *t {
-                Ok(Token::ElementStart(_, _)) => {
-                    self.state = State::Attributes;
-                }
-                Ok(Token::ElementEnd(ref end)) => {
-                    match *end {
-                        ElementEnd::Open => {
-                            self.depth += 1;
-                        }
-                        ElementEnd::Close(_, _) => {
-                            if self.depth > 0 {
-                                self.depth -= 1;
-                            }
-                        }
-                        ElementEnd::Empty => {}
-                    }
-
-                    if self.depth == 0 {
-                        self.state = State::AfterElements;
-                    } else {
-                        self.state = State::Elements;
-                    }
-                }
-                Ok(Token::DtdStart(_, _)) => {
-                    self.state = State::Dtd;
-                }
-                Ok(Token::DtdEnd) => {
-                    self.state = State::Document;
-                }
-                Err(_) => {
-                    self.stream.jump_to_end();
-                    self.state = State::Finished;
-                }
-                _ => {}
-            }
-        }
-
-        t
-    }
+/// Shorthand for:
+///
+/// ```no_run
+/// let start = stream.pos() - 2; // or any other number
+/// some_func().map_err(|e|
+///     Error::InvalidToken(Token::SomeToken, stream.gen_error_pos_from(start), Some(e))
+/// )
+/// ```
+macro_rules! map_err_at {
+    ($fun:expr, $token:expr, $stream:expr, $d:expr) => {{
+        let mut start = $stream.pos() as isize + $d;
+        debug_assert!(start >= 0);
+        if start < 0 { start = 0; }
+        $fun.map_err(|e|
+            Error::InvalidToken($token, $stream.gen_error_pos_from(start as usize), Some(e))
+        )
+    }}
 }
 
 impl<'a> Tokenizer<'a> {
+    /// Enables document fragment parsing.
+    ///
+    /// By default, `xmlparser` will check for DTD, root element, etc.
+    /// But if we have to parse an XML fragment, it will lead to an error.
+    /// This method switch the parser to the root element content parsing mode.
+    /// So it will treat any data as a content of the root element.
+    pub fn set_fragment_mode(&mut self) {
+        self.state = State::Elements;
+        self.is_fragment_parsing = true;
+    }
+
     fn parse_next_impl(s: &mut Stream<'a>, state: State) -> Option<Result<Token<'a>>> {
         if s.at_end() {
             return None;
@@ -213,7 +188,7 @@ impl<'a> Tokenizer<'a> {
                       TokenType::ElementDecl
                     | TokenType::NotationDecl
                     | TokenType::AttlistDecl => {
-                        if let Err(_) = Self::consume_decl(s) {
+                        if Self::consume_decl(s).is_err() {
                             gen_err!(token_type);
                         }
 
@@ -242,7 +217,6 @@ impl<'a> Tokenizer<'a> {
             }
             State::Elements => {
                 let token_type = parse_token_type!();
-
                 match token_type {
                     TokenType::ElementStart => {
                         Self::parse_element_start(s)
@@ -269,7 +243,8 @@ impl<'a> Tokenizer<'a> {
             }
             State::Attributes => {
                 Self::consume_attribute(s).map_err(|e|
-                    Error::InvalidTokenWithCause(TokenType::Attribute, s.gen_error_pos_from(start), e))
+                    Error::InvalidToken(TokenType::Attribute,
+                                        s.gen_error_pos_from(start), Some(e)))
             }
             State::AfterElements => {
                 let token_type = parse_token_type!();
@@ -395,11 +370,7 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_declaration(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 6;
-
-        Self::parse_declaration_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::XMLDecl, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_declaration_impl(s), TokenType::XMLDecl, s, -6)
     }
 
     fn parse_declaration_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
@@ -426,10 +397,6 @@ impl<'a> Tokenizer<'a> {
 
         s.consume_quote()?;
 
-        if ver.to_str() != "1.0" {
-            warn!("Only XML 1.0 is supported.");
-        }
-
         Ok(ver)
     }
 
@@ -443,41 +410,18 @@ impl<'a> Tokenizer<'a> {
 
         s.consume_eq()?;
         s.consume_quote()?;
-        let name = Self::parse_encoding_name(s)?;
-        s.consume_quote()?;
-
-        // TODO: should we keep it?
-        // match name.slice() {
-        //     "UTF-8" | "utf-8" => {}
-        //     _ => {
-        //         // It's not an error, since we already know that input data
-        //         // is a valid UTF-8 string.
-        //         // Probably, the input data has non-UTF-8 encoding, but do not
-        //         // use any of non-Latin-1 characters.
-        //         warn!("Unsupported encoding: {}", name);
-        //     }
-        // }
-
-        Ok(Some(name))
-    }
-
-    // [A-Za-z] ([A-Za-z0-9._] | '-')*
-    fn parse_encoding_name(s: &mut Stream<'a>) -> StreamResult<StrSpan<'a>> {
-        // if !s.at_end() {
-        //     let c = s.curr_byte_unchecked();
-        //     if !Stream::is_letter(c) {
-        //         return Err(XmlXmlError::new(ErrorKind::String("Invalid encoding name".into()),
-        //                                  Some(s.gen_error_pos())));
-        //     }
-        // }
-
-        Ok(s.consume_bytes(|_, c| {
+        // [A-Za-z] ([A-Za-z0-9._] | '-')*
+        // TODO: check that first byte is [A-Za-z]
+        let name = s.consume_bytes(|_, c| {
                c.is_xml_letter()
             || c.is_xml_digit()
             || c == b'.'
             || c == b'-'
             || c == b'_'
-        }))
+        });
+        s.consume_quote()?;
+
+        Ok(Some(name))
     }
 
     // S 'standalone' Eq (("'" ('yes' | 'no') "'") | ('"' ('yes' | 'no') '"'))
@@ -497,10 +441,9 @@ impl<'a> Tokenizer<'a> {
         match value.to_str() {
             "yes" | "no" => {}
             _ => {
-                let c = value.to_str().chars().next().unwrap();
+                let values = vec![value.to_str().into(), "yes".into(), "no".into()];
                 let pos = s.gen_error_pos_from(start);
-                let kind = StreamError::InvalidChar(c, "yn".into(), pos);
-                return Err(StreamError::from(kind).into());
+                return Err(StreamError::InvalidString(values, pos));
             }
         }
 
@@ -527,28 +470,19 @@ impl<'a> Tokenizer<'a> {
 
         if text.to_str().contains("--") {
             let pos = s.gen_error_pos_from(start);
-            return Err(Error::InvalidToken(TokenType::Comment, pos));
+            return Err(Error::InvalidToken(TokenType::Comment, pos, None));
         }
 
         if s.skip_string(b"-->").is_err() {
             let pos = s.gen_error_pos_from(start);
-            return Err(Error::InvalidToken(TokenType::Comment, pos));
+            return Err(Error::InvalidToken(TokenType::Comment, pos, None));
         }
 
-        // if text.slice().ends_with("-") {
-        //     warn!("Comment should not end with '--->'. Last '-' is removed.");
-        //     text = text.slice_region(0, text.len() - 1);
-        // }
-
         Ok(Token::Comment(text))
     }
 
     fn parse_pi(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 2;
-
-        Self::parse_pi_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::PI, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_pi_impl(s), TokenType::PI, s, -2)
     }
 
     // PI       ::= '<?' PITarget (S (Char* - (Char* '?>' Char*)))? '?>'
@@ -582,11 +516,7 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_doctype(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 9;
-
-        Self::parse_doctype_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::DoctypeDecl, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_doctype_impl(s), TokenType::DoctypeDecl, s, -9)
     }
 
     // doctypedecl ::= '<!DOCTYPE' S Name (S ExternalID)? S? ('[' intSubset ']' S?)? '>'
@@ -638,11 +568,7 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_entity_decl(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 8;
-
-        Self::parse_entity_decl_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::EntityDecl, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_entity_decl_impl(s), TokenType::EntityDecl, s, -8)
     }
 
     // EntityDecl  ::= GEDecl | PEDecl
@@ -654,7 +580,6 @@ impl<'a> Tokenizer<'a> {
         let is_ge = if s.curr_byte()? == b'%' {
             s.consume_byte(b'%')?;
             s.consume_spaces()?;
-
             false
         } else {
             true
@@ -693,7 +618,7 @@ impl<'a> Tokenizer<'a> {
                             s.skip_string(b"NDATA")?;
                             s.consume_spaces()?;
                             s.skip_name()?;
-                            warn!("NDataDecl is not supported.")
+                            // TODO: NDataDecl is not supported
                         }
                     }
 
@@ -703,9 +628,9 @@ impl<'a> Tokenizer<'a> {
                 }
             }
             _ => {
+                let chars = vec![c, b'"', b'\'', b'S', b'P'];
                 let pos = s.gen_error_pos();
-                let kind = StreamError::InvalidChar(c as char, "\"'SP".into(), pos);
-                Err(StreamError::from(kind).into())
+                Err(StreamError::InvalidChar(chars, pos))
             }
         }
     }
@@ -714,36 +639,29 @@ impl<'a> Tokenizer<'a> {
         s.consume_spaces()?;
         s.skip_bytes(|_, c| c != b'>');
         s.consume_byte(b'>')?;
-
-        // warn!("'{}' skipped.", s.slice_back(start));
-
         Ok(())
     }
 
+    fn parse_cdata(s: &mut Stream<'a>) -> Result<Token<'a>> {
+        map_err_at!(Self::parse_cdata_impl(s), TokenType::CDSect, s, -9)
+    }
+
     // CDSect  ::= CDStart CData CDEnd
     // CDStart ::= '<![CDATA['
     // CData   ::= (Char* - (Char* ']]>' Char*))
     // CDEnd   ::= ']]>'
-    fn parse_cdata(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 9;
-
+    fn parse_cdata_impl(s: &mut Stream<'a>) -> StreamResult<Token<'a>> {
         let text = s.consume_bytes(|s, c| {
             !(c == b']' && s.starts_with(b"]]>"))
         });
 
-        s.skip_string(b"]]>").map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::CDSect, s.gen_error_pos_from(start), e)
-        )?;
+        s.skip_string(b"]]>")?;
 
         Ok(Token::Cdata(text))
     }
 
     fn parse_element_start(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 1;
-
-        Self::parse_element_start_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::ElementStart, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_element_start_impl(s), TokenType::ElementStart, s, -1)
     }
 
     // '<' Name (S Attribute)* S? '>'
@@ -753,11 +671,7 @@ impl<'a> Tokenizer<'a> {
     }
 
     fn parse_close_element(s: &mut Stream<'a>) -> Result<Token<'a>> {
-        let start = s.pos() - 2;
-
-        Self::parse_close_element_impl(s).map_err(|e|
-            Error::InvalidTokenWithCause(TokenType::ElementClose, s.gen_error_pos_from(start), e)
-        )
+        map_err_at!(Self::parse_close_element_impl(s), TokenType::ElementClose, s, -2)
     }
 
     // '</' Name S? '>'
@@ -793,11 +707,6 @@ impl<'a> Tokenizer<'a> {
         let quote = s.consume_quote()?;
         let value = s.consume_bytes(|_, c| c != quote);
 
-        // if s.curr_byte()? == b'<' {
-        //     let kind = StreamErrorKind::InvalidChar('<', "Char".into(), s.gen_error_pos());
-        //     return Err(StreamXmlError::from(kind).into());
-        // }
-
         s.consume_byte(quote)?;
         s.skip_ascii_spaces();
 
@@ -807,7 +716,7 @@ impl<'a> Tokenizer<'a> {
     fn parse_text(s: &mut Stream<'a>) -> Result<Token<'a>> {
         let text = s.consume_bytes(|_, c| c != b'<');
 
-        let mut ts = Stream::from_span(text);
+        let mut ts = Stream::from(text);
         // TODO: optimize
         ts.skip_spaces();
         if ts.at_end() {
@@ -817,3 +726,56 @@ impl<'a> Tokenizer<'a> {
         }
     }
 }
+
+impl<'a> Iterator for Tokenizer<'a> {
+    type Item = Result<Token<'a>>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        if self.stream.at_end() || self.state == State::Finished {
+            self.state = State::Finished;
+            return None;
+        }
+
+        let t = Self::parse_next_impl(&mut self.stream, self.state);
+
+        if let Some(ref t) = t {
+            match *t {
+                Ok(Token::ElementStart(_, _)) => {
+                    self.state = State::Attributes;
+                }
+                Ok(Token::ElementEnd(ref end)) => {
+                    match *end {
+                        ElementEnd::Open => {
+                            self.depth += 1;
+                        }
+                        ElementEnd::Close(_, _) => {
+                            if self.depth > 0 {
+                                self.depth -= 1;
+                            }
+                        }
+                        ElementEnd::Empty => {}
+                    }
+
+                    if self.depth == 0 && !self.is_fragment_parsing {
+                        self.state = State::AfterElements;
+                    } else {
+                        self.state = State::Elements;
+                    }
+                }
+                Ok(Token::DtdStart(_, _)) => {
+                    self.state = State::Dtd;
+                }
+                Ok(Token::DtdEnd) => {
+                    self.state = State::Document;
+                }
+                Err(_) => {
+                    self.stream.jump_to_end();
+                    self.state = State::Finished;
+                }
+                _ => {}
+            }
+        }
+
+        t
+    }
+}
